#! /usr/bin/python3
# Copyright (c) 2024-2025 National Fundamental Software of China Co., Ltd.
#
# You should have find a copy of NFSChina License. If not, please contact us
# by email <os_support@nfschina.com>

"""
方德进程动态节制(npdm)
====================
动态地检测系统的CPU,IO 和 用户操作事件,提供2个功能:
静置冻结: 当用户不再操作电脑时，冻结黑名单上的程序。
动态节制: 当CPU/IO压力过大时,把占用高的进程节制起来。

配置文件:/etc/default/npdm
日志文件:/var/log/npdm.log
PID文件:/tmp/npdm.pid

Copyright (c) 2024 Li Qiong <liqong@nfschina.com>
"""

import os
import sys
import time
import threading
import signal

from pdm.procinfo import (procs_load,get_pids_by_names)
from pdm.sysevent import SysEvent
from pdm.procontr import ProcControl
from pdm.config   import Configure
from pdm.log      import logger

pid_file = "/tmp/npdm.pid"

status = {"exit": 0}

#pid_control = {"freeze":[], "moderate":[], "lowactive":[]}

def sleep_check(interval=5):
    for i in range(interval):
        if status["exit"] == 1:
            break
        time.sleep(1)

"""
静置冻结
=======
    FreezeEnable:     "1"-启动, "0"-关闭.
    FreezeOnBattery:  "1"-使用电池时生效. "0"-总是有效.
    InactivityTime:   多长时间没有操作被认为是静置状态,默认是150秒.
"""
def noact_loop(config, sysev, procntr):
    if not config.noAct["Freeze"] and not config.noAct["LowActive"]:
        logger.i(f"[noAct] no app list, exit")
        return

    last = 0
    while True:
        if config.noAct["OnBattery"] and sysev.checkBattery() == 0:
            time.sleep(300)
            continue

        act = sysev.checkUserActivity(config.noAct["Timeout"])
        if act == SysEvent.EVT_USR_NOACT:
            if last == SysEvent.EVT_USR_NOACT:
                continue; 

            last = SysEvent.EVT_USR_NOACT
            pids = get_pids_by_names(config.noAct["Freeze"])
            for pid in pids:
                procntr.freeze(pid)
            logger.d(f"[noAct] freeze {pids}")

            pids = get_pids_by_names(config.noAct["LowActive"])
            for pid in pids:
                procntr.lowerActiveByNoAct(pid)
            logger.d(f"[noAct] lowerActive {pids}")

        elif act == SysEvent.EVT_USR_ACT:
            if last == SysEvent.EVT_USR_NOACT:
                last = SysEvent.EVT_USR_ACT
                procntr.unfreezeAll()
                procntr.unnoActAll()
                logger.d("[noAct] free all pids")

            time.sleep(5) # May user keep action, moderate check


def lowactive_loop(config, procntr):
    if not config.lowActive["BlackList"]:
        logger.i(f"[lowactive]] no blacklist, exit")
        return

    while True:
        pids = get_pids_by_names(config.lowActive["BlackList"])
        for pid in pids:
            procntr.lowerActive(pid)
        logger.d(f"[lowactive]  lowerActive {pids}")
        time.sleep(600)

# The main_loop running forevery, forc exit by SIGTERM
def exit_main_loop(procntr):
    procntr.unfreezeAll()
    procntr.unnoActAll()
    procntr.unlowerActiveAll()
    os.remove(pid_file)
    logger.i("npdm service stop ..... Bye")

def main_loop(config, sysev, procntr):
    noActThd = None

    if config.noActEnable:
        logger.i(f"npdm start noAct proc ...\n")
        noActThd = threading.Thread(target=noact_loop, args=(config, sysev,procntr))
        noActThd.start()

    if config.lowActiveEnable:
        logger.i(f"npdm start lowactive ...\n")
        lowactive_loop(config, procntr)

    # Not running here, thread running forevery, abort by signal.
    if noActThd != None:
        noActThd.join()

def handle_signal(signum, frame, procntr):
    exit_main_loop(procntr)
    os._exit(0)

if __name__ == '__main__':
    if os.path.exists(pid_file):
        with open(pid_file, 'r') as f:
            pid = f.read().strip()
            logger.i(f"Daemon already running with PID: {pid}")
            sys.exit(1)

    scriptdir =  os.path.dirname(os.path.abspath(__file__))
    #print(f"{scriptdir}")
    os.chdir(scriptdir)
    with open(pid_file, 'w') as f:
        f.write(str(os.getpid()) + '\n')

    logger.i(f"npdm service start ..... Hello \n")

    config = Configure()
    sysev = SysEvent()
    procntr = ProcControl(config)

    signal.signal(signal.SIGINT, lambda signum, frame: handle_signal(signum, frame, procntr))

    logger.setLevel(config.logLevel)

    logger.i(f"npdm service enter main_loop \n")

    main_loop(config, sysev, procntr)

    # Not running here
    exit_main_loop(procntr)
