#! /usr/bin/python3
# Scan /proc, parse process status
#
# Copyright (c) 2024 Li Qiong <liqong@nfschina.com> 
# Copyright (c) 2024-2025 National Fundamental Software of China Co., Ltd.
#
# You should have find a copy of NFSChina License. If not, please contact us
# by email <os_support@nfschina.com>

import os
import time
from collections import defaultdict

from pdm.sysevent import SysEvent
from pdm.log      import logger

MANAGE_PID_START=500

class ProcessInfo:
    def __init__(self, pid):
        self.pid = pid
        self.uid = 0
        self.runtime = 0
        self.rbio = 0
        self.wbio = 0
        self.rio = 0
        self.wio = 0
        self.iototal = 0
        self.biototal = 0
        self.name=""

    @classmethod
    def from_pid(cls, pid):
        proc = cls(pid)
        proc.cpuTime(pid)
        proc.ioStat(pid)
        proc.procName(pid)
        return proc

    def dump(self):
        print(f"name:{self.name}, {self.pid} --> time: {self.runtime} IO: {self.iototal}, BIO: {self.biototal}")

    def procName(self, pid):
        with open(f"/proc/{pid}/comm", "r") as f:
            global nmtopid
            self.name = f.read().strip()

    def cpuTime(self, pid):
        with open(f"/proc/{pid}/stat", "r") as f:
            fields = f.readline().split()
            utime = int(fields[13])  # User mode time in clock ticks
            stime = int(fields[14])  # Kernel mode time in clock ticks
            #cutime = int(fields[15]) # User mode time of children in clock ticks
            #cstime = int(fields[16]) # Kernel mode time of children in clock ticks
            self.runtime = utime + stime

    def ioStat(self, pid):
        with open(f"/proc/{pid}/io", "r") as f:
            io_data = f.readlines()
            for line in io_data:
                key, value = line.split(":")
                if key  == "rchar":
                    self.rio = int(value)
                elif key == "wchar":
                    self.wio = int(value)
                elif key == "read_bytes":
                    self.rbio = int(value)
                elif key == "write_bytes":
                    self.wbio = int(value)
            self.iototal = self.rio + self.wio
            self.biototal = self.rbio + self.wbio
            #print(f"PID:{pid} --> IO:{self.iototal}, BIO:{self.biototal}")

def is_target(entry):
    with open(os.path.join("/proc", entry, "comm"), "r") as f:
        name = f.read().strip()
        return not name.startswith('[')

""" procname: [pid, pid]"""
def procname_list():
    procs = defaultdict(list)
    for entry in os.listdir("/proc"):
        if not entry.isdigit():
            continue
        pid = int(entry)
        if (pid > MANAGE_PID_START):
            try:
                if not is_target(entry):
                    continue
                p = ProcessInfo(pid)
                p.procName(pid)
                if p.name != "":
                    procs[p.name].append(pid)
            except:
                continue
    return procs

def get_pids_by_names(names):
    ret = []
    procs = procname_list()
    for nm in names:
        pids = procs.get(nm)
        if pids is not None:
            ret.extend(pids)
    return ret


def proclist():
    processes = {}
    total_io = 0
    for entry in os.listdir("/proc"):
        if not entry.isdigit():
            continue
        pid = int(entry)
        if (pid > MANAGE_PID_START):
            try:
                if not is_target(entry):
                    continue
                procInfo = ProcessInfo.from_pid(pid)
                total_io += procInfo.iototal
                processes[pid] = procInfo
            except:
                continue
    return total_io,processes

def cpu_total_times():
    with open("/proc/stat", "r") as f:
        line = f.readline() 
    fields = line.strip().split()[1:]
    return sum(list(map(int, fields)))

def procs_load(type=3, interval=1.0):
    total_cpu_1 = cpu_total_times()
    total_io_1,pl1 = proclist()

    time.sleep(interval)

    total_cpu_2 = cpu_total_times()
    total_io_2,pl2 = proclist()

    cpu_amount = total_cpu_2 - total_cpu_1
    io_amount = total_io_2 - total_io_1
    #print(f"cpu time {cpu_amount}, io bytes: {io_amount}")
    if cpu_amount <= 0 or io_amount <= 0:
        return []

    load = []
    for pid in pl2:
        if pid not in pl1:
            continue
        if (type & SysEvent.EVT_CPU_LOAD):
            cpu_usage = ((pl2[pid].runtime - pl1[pid].runtime)  / cpu_amount) * 100
            load.append((pid, cpu_usage))
            logger.d(f"[procs_load]: {pl2[pid].name} {pid} CPU:{cpu_usage}")
        else:
            io_usage = ((pl2[pid].iototal - pl1[pid].iototal) / io_amount) * 100
            load.append((pid, io_usage))
            logger.d(f"[procs_load]: {pl2[pid].name} {pid} IO:{io_usage}")

    load.sort(key=lambda x: x[1], reverse=True)
    load = load[:4]
    max = load[0][1]
    #print(f"max:{max}: {load}")
    load = [t for t in load if (max - t[1]) < 0.5]
    pids = [t[0] for t in load]
    return pids

def monitor_cpu_io(interval=5.0):
    print(f"==========================================")
    while True:
        io_total,processes = proclist()
        print(f"io_total:{io_total} bytes, cpu_total:{cpu_total_times()} jiffies")
        for pid in processes:
            proc = processes[pid]
            print(f"name:{proc.name}, {proc.pid:<6} --> time: {proc.runtime:<6} IO: {proc.iototal:<10}, BIO: {proc.biototal:<10}")
        time.sleep(interval)

def dump_procname(interval=5.0):
    print(f"==========================================")
    processes = procname_list()
    print(f"{len(processes)}")
    for name in processes:
        pid = processes[name]
        print(f"name:{name}, {pid}")

if __name__ == '__main__':
    #monitor_cpu_io()
    #procInfo = ProcessInfo.from_pid(3567)
    #procInfo.dump()
    #dump_procname()
    names=["wps", "et", "sd_dummysss"]
    pids = get_pids_by_names(names)
    print(pids)
