# Check system event: cpu , io load, battery
#
# Copyright (c) 2024 Li Qiong <liqong@nfschina.com> 
# Copyright (c) 2024-2025 National Fundamental Software of China Co., Ltd.
#
# You should have find a copy of NFSChina License. If not, please contact us
# by email <os_support@nfschina.com>

import time
import select
import os

from pdm.log  import logger

CPU_LOAD_THRES=95
IO_WAIT_THRES=40
IO_STALL_THRES=10

PRERSURE_IO_THRES=0.5
PRERSURE_CPU_THRES=30

class SysEvent:
    EVT_CPU_LOAD=0x01
    EVT_IO_LOAD=0x02
    EVT_USR_ACT=0x03
    EVT_USR_NOACT=0x04

    def  __init__(self):
        self.cpuload = 0
        self.iowait = 0
        self.iostall = 0
        self.battery = 0

    def dump(self, type, stats):
        print(f"{type} Pressure Statistics:")
        for key, value in stats.items():
            print(f"{key}: {value}")
            print("-" * 40)

    def getCpuTimes(self):
        with open("/proc/stat", "r") as f:
            first_line = f.readline()
            fields = first_line.strip().split()[1:]
        return list(map(int, fields))

    def calcCpuLoad(self,interval=5):
        times1 = self.getCpuTimes()
        time.sleep(interval)
        times2 = self.getCpuTimes()
        deltas = [t2 - t1 for t1, t2 in zip(times1, times2)]
        idle_delta = deltas[3]  # idle is the 4th field
        total_delta = sum(deltas)

        self.cpuload = 100 * (1 - idle_delta / total_delta)
        self.iowait = deltas[4]

    def getPressure(self, type):
        try:
            stats = {}
            with open(f"/proc/pressure/{type}", "r") as f:
                for line in f:
                    parts = line.split()
                    category = parts[0]
                    values = {k: float(v) for k, v in (item.split("=") for item in parts[1:])}
                    stats[category] = values
            return stats
        except  Exception as e:
            logger.e(f"[getPressure] can't read /proc/pressure/io {e}")
            return {}

    def checkPressure(self, stallThres):
        trig = f"some {stallThres * 1000000} 5000000\n"
        pressure_dir = "/proc/pressure"
        fds=[]
        try:
            files = [os.path.join(pressure_dir, f) for f in ["io", "cpu"] ]
            fds = [os.open(f, os.O_RDWR | os.O_NONBLOCK) for f in files]
            poller = select.poll()
            for fd in fds:
                os.write(fd, trig.encode('utf-8'))
                poller.register(fd, select.POLLPRI)
            events = poller.poll()  # Timeout in milliseconds
            logger.i(f"[checkPressure]: events:{events}, fds:{fds}, trig:{trig}")

            for fd,e in events:
                print("f{e} {fd}")
                if e&select.POLLPRI:
                    if fd == fds[0]:
                        return SysEvent.EVT_IO_LOAD
                    return SysEvent.EVT_CPU_LOAD

        except Exception as e:
            logger.e(f"[checkUserActivity] error: {e}")
        finally:
            for fd in fds:
                os.close(fd)
        return 0

    def checkUserActivity(self, timeout=150):
        input_dir = "/dev/input"
        fds=[]
        try:
            devices = [os.path.join(input_dir, f) for f in os.listdir(input_dir) if f.startswith("event")]
            fds = [os.open(dev, os.O_RDONLY) for dev in devices]
            #print(f"Monitoring user activity on {len(fds)} devices for {timeout} seconds...")
            poller = select.poll()
            for fd in fds:
                poller.register(fd, select.POLLIN)

            events = poller.poll(timeout * 1000)  # Timeout in milliseconds
            if events:
                return SysEvent.EVT_USR_ACT
            else:
                return SysEvent.EVT_USR_NOACT

        except Exception as e:
            logger.e(f"[checkUserActivity] error: {e}")
        finally:
            for fd in fds:
                os.close(fd)

    def checkBattery(self):
        power_supply_path = "/sys/class/power_supply/"
        battery_dirs = [d for d in os.listdir(power_supply_path) if d.startswith("BAT")]

        if not battery_dirs:
            return 0

        # Check the first battery's status
        battery_status_path = os.path.join(power_supply_path, battery_dirs[0], "status")
        try:
            with open(battery_status_path, "r") as status_file:
                status = status_file.read().strip()
                if status == "Discharging":
                    self.battery = 1
                else:
                    self.battery = 0
        except Exception as e:
            logger.e(f"checkBattery error: {e}")

        return self.battery 

    def monitorPressure(self):
        while True:
            ret = self.getPressure("io")
            print(f"{ret['some']['avg10']}")
            ret = self.getPressure("cpu")
            print(f"{ret['some']['avg10']}")
            time.sleep(5)
            print("==================")


if __name__ == '__main__':
    #while True:
    #    monitor_user_activity(timeout=10)
    sysevt = SysEvent()
